﻿import { expressionPart } from './parts/expressionPart';
import { nullPart } from './parts/nullPart';
import { selfPart } from './parts/selfPart';
import { Stack } from './../utils/stack';
import { StringBuilder } from 'typescript-dotnet-commonjs/System/Text/StringBuilder';
import List from "typescript-dotnet-commonjs/System/Collections/List";
import { operationPart } from './parts/operationPart';
import { partList } from './parts/partList';
import { indicatorPart } from './parts/indicatorPart';
import { literalPart } from './parts/literalPart';
import { propertyPart } from './parts/propertyPart';
import { bracketPart } from './parts/bracketPart';
import { indexerPart } from './parts/indexerPart';
import { functionCallPart } from './parts/functionCallPart';
import { switchCasePart } from './parts/switchCasePart';
import { aggregateFunctionCallPart } from './parts/aggregateFunctionCallPart';
import { utils } from "./utils";
import { unterminatedStringLiteralException } from "../error/unterminatedStringLiteralException";
import { unterminatedCharacterLiteralException } from "../error/unterminatedCharacterLiteralException";
import { unknownFunctionException } from "../error/unknownFunctionException";
import { badNumericLiteralException } from "../error/badNumericLiteralException";
import { badCharacterLiteralException } from "../error/badCharacterLiteralException";
import { badExpressionException } from "../error/badExpressionException";
import { functionsList } from "../functions/functionsList";
import { operatorsList } from "../operators/operatorsList";
import { aggregateFunction } from "../functions/aggregateFunction";
import { mfunction } from "../functions/mfunction";
import { outputVariables } from "../utils/outputVariables";
import { isNumberType } from "../utils/general";

export interface ICompileFlags {
    legacyMode: boolean;
    aggregationSection: boolean;
    bracketPart: boolean;
    indexerPart: boolean;
    switchCasePart: boolean;
}

export class compileFlags {
    private _legacyMode: boolean = false;
     get legacyMode(): boolean {
        return this._legacyMode;
    }
     set legacyMode(value: boolean) {
        this._legacyMode = value;
    }

    private _aggregationSection: boolean = false;

     get aggregationSection(): boolean {
        return this._aggregationSection;
    }

     set aggregationSection(value: boolean) {
        this._aggregationSection = value;
    }
    private _bracketPart: boolean = false;
     get bracketPart(): boolean {
        return this._bracketPart;
    }
     set bracketPart(value: boolean) {
        this._bracketPart = value;
    }

    private _indexerPart: boolean = false;
     get indexerPart(): boolean {
        return this._indexerPart;
    }
     set indexerPart(value: boolean) {
        this._indexerPart = value;
    }

    private _switchCasePart: boolean = false;
     get switchCasePart(): boolean {
        return this._switchCasePart;
    }
     set switchCasePart(value: boolean) {
        this._switchCasePart = value;
    }

     constructor(original?: ICompileFlags) {
        if (original != undefined) {
            this.legacyMode = original.legacyMode;
            this.aggregationSection = original.aggregationSection;
            this.bracketPart = original.bracketPart;
            this.indexerPart = original.indexerPart;
            this.switchCasePart = original.switchCasePart;
        }
    }
}

export class expressionCompiler {

     static ProcessLegacy(expression: string): string {
        expression = expression.replace('Format\(([^""][^\)]+[^""])\)', 'Format("{$1}")');
        return expression;
    }

     static compile(path: string, startIndex: number | undefined, endIndex: number | undefined, flags: compileFlags): expressionPart | undefined {
        if (startIndex === undefined) {
            if (path == null) {
                return undefined;
            }

            return expressionCompiler.compile(path, 0, (path.length - 1), flags);
        }
        else {
            if (path.trim() === ".") {
                return new selfPart();
            }

            let sbPart: StringBuilder = new StringBuilder();
            let currentPart: expressionPart | undefined = undefined;
            let postFixList: expressionPart[] = [];
            let operationsStack: Stack<operationPart> = new Stack<operationPart>();
            let i: number = startIndex;
            let isNumber: boolean = false;
            let isFloatingPoint: boolean = false;
            let newExpression: boolean = true;
            let inString: boolean = false;
            let inChar: boolean = false;
            let isDot: boolean = false;
            let isIndicator: boolean = false;
            let escapeNext: boolean = false;
            let stringChar: string = "\"";
            const StartNewPart = (isNewExpression: boolean) => {
                newExpression = isNewExpression ? true : newExpression;
                isNumber = false;
                isFloatingPoint = false;
                isDot = false;
            };

            const AddNewPart = (nextPart: expressionPart) => {
                if (nextPart == undefined)
                    return;

                if (currentPart instanceof partList) {
                    (currentPart as partList).addSubPart(nextPart);
                }
                else if (currentPart == undefined) {
                    currentPart = nextPart;
                }
                else {
                    let ep: partList = new partList();

                    ep.addSubPart(currentPart);
                    ep.addSubPart(nextPart);

                    currentPart = ep as expressionPart;
                }
            };
            const MarkEnd = (final: boolean) => {
                let nextPart: expressionPart | undefined = undefined;

                let val: string = sbPart.toString() == "\"" ? "" : sbPart.toString();

                sbPart.clear();

                if (isIndicator) {
                    nextPart = new indicatorPart();
                    (nextPart as indicatorPart).value = val;

                    isIndicator = false;
                }
                else if (inString) {
                    nextPart = new literalPart();
                    (nextPart as literalPart).value = val;

                }
                else if (val.length > 0) {
                    if (inChar) {
                        nextPart = new literalPart();
                        (nextPart as literalPart).value = val[0];
                    }
                    else if (val.equals("true") || val.equals("false")) {
                        nextPart = new literalPart();
                        (nextPart as literalPart).value = val.equals("true");
                    }
                    else if (val.equals("null")) {
                        nextPart = new nullPart();
                    }
                    else if (isNumber) {
                        try {
                            if (val.equals(".")) {
                                nextPart = new selfPart();
                            }
                            else {
                                let numeric: Object = 0;

                                if (val.endsWith("f") || val.endsWith("d")) {
                                    isFloatingPoint = true;
                                }

                                if (isFloatingPoint) {
                                    if (val.endsWith("f") || val.endsWith("F"))
                                        numeric = Number.parseFloat(val.substring(0, val.length - 1));
                                    else if (val.endsWith("d") || val.endsWith("D"))
                                        numeric = Number.parseFloat(val.substring(0, val.length - 1));
                                    else
                                        numeric = Number.parseFloat(val);
                                }
                                else {
                                    if (val.endsWith("l") || val.endsWith("L"))
                                        numeric = Number.parseInt(val.substring(0, val.length - 1));
                                    else
                                        numeric = Number.parseInt(val);
                                }

                                nextPart = new literalPart();
                                (nextPart as literalPart).value = numeric;
                            }
                        }
                        catch (ex) {
                            throw new badNumericLiteralException(val);
                        }
                    }
                    else {
                        nextPart = new propertyPart();
                        (nextPart as propertyPart).propertyName = val;
                    }
                }

                AddNewPart(nextPart as expressionPart);

                if (final && currentPart != null) {
                    postFixList.push(currentPart);
                    currentPart = undefined;
                }

                isDot = false;
            };
            for (let i: number = startIndex; (i <= (endIndex as number)); i++) {
                let c: string = path[i];
                let currentOperator: string | null = null;
                if (flags.switchCasePart && newExpression && c == "#") {
                    isIndicator = true;
                }
                else if (((c == stringChar && !inChar) || (flags.legacyMode && c == '\'' && (!inString || stringChar == '\''))) && !escapeNext) {

                    newExpression = false;
                    isNumber = false;
                    isFloatingPoint = false;
                    if (c == '\'')
                        stringChar = '\'';

                    if (inString) {
                        MarkEnd(false);
                        StartNewPart(false);
                        stringChar = '"';
                    }

                    inString = !inString;
                }
                else if (inString && !inChar) {
                    newExpression = false;
                    if (c == '\\' && !escapeNext) {
                        escapeNext = true;
                    }
                    else {
                        sbPart.append(c);
                        escapeNext = false;
                    }

                }
                else if (c == '\'' && !escapeNext && !inString) {
                    newExpression = false;
                    isNumber = false;
                    isFloatingPoint = false;
                    if (inChar) {
                        MarkEnd(false);
                        StartNewPart(false);
                    }

                    inChar = !inChar;
                }
                else if ((inChar && !inString)) {
                    newExpression = false;
                    if (c == '\\' && !escapeNext) {
                        escapeNext = true;
                    }
                    else {
                        if ((sbPart.toString().length < 1)) {
                            sbPart.append(c);
                        }
                        else {
                            throw new badCharacterLiteralException(sbPart.toString());
                        }

                        escapeNext = false;
                    }

                }
                else if (!inString && !inChar) {
                    if (String.isWhiteSpace(c)) {
                        if ((sbPart.toString().length > 0)) {
                            MarkEnd(true);
                        }

                        StartNewPart(false);
                        continue;
                    }
                    else if (newExpression && !isDot && (isNumberType(c) || c == '-' || c == '.')) {
                        newExpression = false;
                        isNumber = true;
                        isFloatingPoint = c == '.';
                        sbPart.append(c);
                    }
                    else if (isNumber && !isDot && (isNumberType(c) || (c == "." && !isFloatingPoint))) {
                        if ((c == '.')) {
                            isFloatingPoint = true;
                        }

                        sbPart.append(c);
                    }
                    else if (operatorsList.binaryOperators.containsKey(c + "") && !String.isNullOrEmpty(currentOperator = c + "")) {
                        MarkEnd(true);
                        if ((flags.legacyMode
                            && (currentOperator == "%"))) {
                            currentOperator = "#";
                        }

                        let op: operationPart = new operationPart();
                        op.operator = operatorsList.binaryOperators.getValue(currentOperator);
                        while (((operationsStack.count() > 0) && ((operationsStack.peek().operator?.precedence as number) >= (op.operator?.precedence as number)))) {
                            let peekPart: operationPart = operationsStack.pop();
                            postFixList.push(peekPart);
                        }

                        operationsStack.push(op);
                        StartNewPart(true);
                    }
                    else if ((c == '{')) {
                        MarkEnd(false);
                        let subEndIndex: outputVariables = new outputVariables();
                        subEndIndex.endIndex = 0;
                        let switchSubject: expressionPart | undefined = currentPart;
                        currentPart = undefined;
                        AddNewPart(expressionCompiler.parseSwitchCasePart(path, i, /* out */subEndIndex, switchSubject, flags));
                        MarkEnd(false);
                        StartNewPart(false);
                        i = subEndIndex.endIndex;
                        newExpression = false;
                    }
                    else if ((c == '(')) {
                        let subEndIndex: outputVariables = new outputVariables();

                        if ((sbPart.toString().length > 0)) {
                            let ep: expressionPart | null = null;
                            ep = expressionCompiler.parseFunctionCallPart(path, i, /* out */subEndIndex, sbPart.toString(), flags);
                            sbPart.clear();
                            AddNewPart(ep);
                        }
                        else if (newExpression) {
                            AddNewPart(expressionCompiler.parseBracketPart(path, i, /* out */subEndIndex, flags));
                        }
                        else {
                            throw new badExpressionException("Found '(' where expected property");
                        }

                        i = subEndIndex.endIndex;
                        newExpression = false;
                    }
                    else if ((c == '[')) {
                        newExpression = false;
                        if ((sbPart.toString().length > 0)) {
                            MarkEnd(false);
                        }

                        // let subEndIndex: number = 0;
                        let subEndIndex: outputVariables = new outputVariables();
                        subEndIndex.endIndex = 0;
                        AddNewPart(expressionCompiler.parseIndexerPart(path, i, subEndIndex, flags));
                        StartNewPart(false);
                        i = subEndIndex.endIndex;
                    }
                    else if (((c == '.')
                        && (!isNumber
                            && !isFloatingPoint))) {
                        if ((sbPart.toString().length > 0)) {
                            MarkEnd(false);
                        }

                        isDot = true;
                        newExpression = false;
                    }
                    else {
                        sbPart.append(c);
                        newExpression = false;
                    }

                }

            }

            if (inString) {
                throw new unterminatedStringLiteralException();
            }
            else if (inChar) {
                throw new unterminatedCharacterLiteralException();
            }

            MarkEnd(true);
            while ((operationsStack.count() > 0)) {
                postFixList.push(operationsStack.pop());
            }

            let result: expressionPart = expressionCompiler.solvePostFix(postFixList);
            if ((result instanceof propertyPart && (flags.indexerPart && flags.legacyMode))) {
                let oldResult: string = (result as propertyPart).propertyName;
                result = new literalPart();
                (result as literalPart).value = oldResult;
            }

            if (flags.bracketPart) {
                let bp: bracketPart = new bracketPart();
                bp.addSubPart(result);
                return bp;
            }
            else {
                return result;
            }
        }
    }

     static solvePostFix(postFixList: expressionPart[]): expressionPart {
        let partStack: Stack<expressionPart> = new Stack<expressionPart>();
        postFixList.forEach((item) => {
            if (!(item instanceof operationPart)) {
                partStack.push(item);
            }
            else {
                let op: operationPart = item as operationPart;
                op.parameter2 = partStack.pop();
                op.parameter1 = partStack.pop();
                partStack.push(op);
            }
        })
        return partStack.peek();
    }

    private static parseBracketPart(path: string, startIndex: number, /* out */endIndex: outputVariables, flags: ICompileFlags): bracketPart {
        endIndex.endIndex = startIndex;
        utils.markEnclosure(path, startIndex, "(", ")", /* out */endIndex);
        let compileFlag = new compileFlags(flags);
        compileFlag.bracketPart = true;
        let bp: bracketPart = (<bracketPart>(expressionCompiler.compile(path, (startIndex + 1), (endIndex.endIndex - 1), compileFlag)));
        bp.BracketEnded = true;
        return bp;
    }

    private static parseFunctionCallPart(path: string, startIndex: number, /* out */endIndex: outputVariables, functionName: string, flags: compileFlags): expressionPart {
        let parameterTokens: List<string> = utils.splitEnclosure(path, startIndex, "(", ")", /* out */endIndex, ",");
        if (functionsList.ExpressionFunctions.containsKey(functionName)) {
            let functionCallPart1: functionCallPart = new functionCallPart();
            functionCallPart1.function = functionsList.ExpressionFunctions.getValue(functionName);
            functionCallPart1.parameters = (functionCallPart1.function as mfunction).compileParameters(parameterTokens, flags);
            return functionCallPart1;
        }
        else if ((flags.aggregationSection && functionsList.AggregateFunctions.containsKey(functionName))) {
            let functionCallPart: aggregateFunctionCallPart = new aggregateFunctionCallPart();
            functionCallPart.Function = functionsList.AggregateFunctions.getValue(functionName) as aggregateFunction;
            functionCallPart.parameters = functionCallPart.Function.compileParameters(parameterTokens, flags);
            return functionCallPart;
        }
        else {
            throw new unknownFunctionException(functionName.toString());
        }

    }

     static parseIndexerPart(path: string, startIndex: number, endIndex: outputVariables, flags: compileFlags): indexerPart {
        let result: indexerPart = new indexerPart();
        utils.markEnclosure(path, startIndex, "[", "]", /* out */endIndex);
        const indexCompileFlag = new compileFlags(flags);
        indexCompileFlag.indexerPart = true;
        result.indexerExpression = expressionCompiler.compile(path, (startIndex + 1), (endIndex.endIndex - 1), indexCompileFlag) as expressionPart;
        return result;
    }

     static parseSwitchCasePart(path: string, startIndex: number, endIndex: outputVariables, expression: expressionPart | undefined, flags: compileFlags): switchCasePart {
        let scp: switchCasePart = new switchCasePart();
        scp.expression = expression as expressionPart;
        scp.switchData = utils.buildSwitchCase(path, startIndex, /* out */endIndex, new compileFlags(flags));
        return scp;
    }
}
