﻿import { mfunction } from './../mfunction'
import expressionData from '../../expressionData'
import { expressionPart } from './../../compiler/parts/expressionPart'
import { List } from 'typescript-dotnet-commonjs/System/Collections/List'
import { groupDataContext } from './../../functions/groupDataContext';
import { KeyValuePair } from 'typescript-dotnet-commonjs/System/KeyValuePair'
import { IEnumerable } from 'typescript-dotnet-commonjs/System/Collections/Enumeration/IEnumerable';
import { literalPart } from '../../compiler/parts/literalPart';
import { DateTime } from 'typescript-dotnet-commonjs/System/Time/DateTime'
import { compileFlags } from '../../compiler/expressionCompiler';
import { getType } from '../../utils/reflectionExtender';
import { typeMapper } from '../../types/typeMapper';
import { namedSelection } from '../namedSelection';
import { outputVariables } from '../../utils/outputVariables';
import { isFloatType, isNumberType } from '../../utils/general';
import { selection } from '../selection';
import { DictionaryObject } from '../../utils/dictionaryObject';
import { parse } from 'date-fns';

export class toList extends mfunction {

    get minParameters(): number {
        return 1;
    }

    get maxParameters(): number {
        return -1;
    }

    compileParameters(paramList: List<string>, flags: compileFlags): expressionPart[] {
        if (flags.legacyMode) {
            let parts: expressionPart[] = [];
            for (let param of paramList.toArray()) {
                let lPart = new literalPart();
                lPart.value = param.trim();
                parts.push(lPart);
            }
            return parts;
        }
        else {
            return super.compileParameters(paramList, flags);
        }

    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object {
        resolved.resolved = true;
        let result: List<Object> = new List<Object>();
        for (let ex of parameters) {
            let str: Object = ex.resolve(expressionData, resolved, running) as Object;
            if (str !== undefined)
                result.add(str);
        }

        resolved.resolved = true;
        return result;
    }
}

export class toDictionary extends mfunction {

     /* override */ get minParameters(): number {
        return 1;
    }

     /* override */ get maxParameters(): number {
        return -1;
    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running?: groupDataContext): Object {
        resolved.resolved = true;
        let result: DictionaryObject = new DictionaryObject();
        for (let ex of parameters) {
            let object;
            let val = (ex.resolveGeneric<DictionaryObject>(
                expressionData,
                resolved,
                getType(new DictionaryObject()),
                running));
            let key = val.getKeys()[0];
            let valueOfDic = val.getValue(key);
            result.add(key, valueOfDic as Object);
        }

        resolved.resolved = true;
        return result;
    }
}

export class toDate extends mfunction {

     /* override */ get minParameters(): number {
        return 1;
    }

     /* override */ get maxParameters(): number {
        return 1;
    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running?: groupDataContext): Object | undefined {
        let value: unknown = null;
        resolved.resolved = true;
        if ((parameters.length > 0)) {
            value = parameters[0].resolve(expressionData, /* out */resolved, /* running:*/ undefined);
        }

        let paramDate: Date;
        if (resolved.resolved && (value != null)) {
            let paramIsDate: boolean = false;
            if ((value instanceof DateTime)) {
                paramDate = (<DateTime>(value)).toJsDate();
                paramIsDate = true;
            }
            else {
                if (typeof value === 'string' && expressionData.dtFormat !== undefined) {
                    try {
                        const sformat = expressionData.dtFormat.replace('0:', '').
                            replace('{', '').
                            replace('}', '').replace('tt', 'aa').replace('dddd', 'EEEE').replace('ddd', 'eee');
                        const dt1 = new Date();
                        parse(value, sformat, dt1);
                        const dt: Date = new Date(Date.parse(dt1.toString()))
                        paramIsDate = true;
                        if (!isNaN(dt.getTime()))
                            return dt;
                    }
                    catch (e) { }
                }

                paramDate = new Date((value as string).toString());
                paramIsDate = true;

            }

            if (paramIsDate) {
                //Object.values[]
                return paramDate.getTime();
            }

        }

        return undefined;
    }
}

export class toNumber extends mfunction {

     /* override */ get minParameters(): number {
        return 1;
    }

     /* override */ get maxParameters(): number {
        return 2;
    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        let value: Object | undefined = undefined;
        resolved.resolved = true;
        let type: string = "auto";
        if ((parameters.length > 0)) {
            value = parameters[0].resolve(expressionData, /* out */resolved, /* running:*/ undefined);
        }

        if ((parameters.length > 1)) {
            type = parameters[1].resolve(expressionData, /* out */resolved, /* running:*/ undefined) as string;
        }

        if ((resolved.resolved
            && (value !== undefined))) {
            switch (type) {
                // case "sbyte":
                //     return typeMapper.toType<System.SByte>(value);
                //     break;
                case "short":
                    return typeMapper.toType<number>(value);
                    break;
                case "int":
                    return typeMapper.toType<number>(value);
                    break;
                case "long":
                    return typeMapper.toType<number>(value);
                    break;
                case "byte":
                    return typeMapper.toType<number>(value);
                    break;
                case "ushort":
                    return typeMapper.toType<number>(value);
                    break;
                case "uint":
                    return typeMapper.toType<number>(value);
                    break;
                case "ulong":
                    return typeMapper.toType<number>(value);
                    break;
                case "float":
                    return typeMapper.toType<number>(value);
                    break;
                case "double":
                    return typeMapper.toType<number>(value);
                    break;
                case "auto":
                default:
                    if (value === null)
                        return 0;
                    if (isNumberType(value) && isFloatType(value as number)) {
                        return Number.parseFloat(value.toString());
                    }
                    else if (isNumberType(value)) {
                        return Number.parseInt(value.toString());
                    }
                    else
                        return value;
                    break;
            }

        }
        else {
            switch (type) {
                case "short":
                    return (<number>(0));
                    break;
                case "int":
                    return (<number>(0));
                    break;
                case "long":
                    return (<number>(0));
                    break;
                case "byte":
                    return (<number>(0));
                    break;
                case "ushort":
                    return (<number>(0));
                    break;
                case "uint":
                    return (<number>(0));
                    break;
                case "ulong":
                    return (<number>(0));
                    break;
                case "float":
                    return (<number>(0));
                    break;
                case "double":
                    return (<number>(0));
                    break;
                case "auto":
                    break;
                default:
                    return 0;
                    break;
            }

        }

    }
}

export class coalesce extends mfunction {

    get minParameters(): number {
        return 2;
    }

    get maxParameters(): number {
        return 3;
    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        if ((parameters.length > 1)) {
            let subResolved: outputVariables = new outputVariables();
            let value: Object | undefined = parameters[0].resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
            let emptyAsNull = parameters.length > 2 ? parameters[2].resolveGeneric<boolean>(expressionData, subResolved, getType(Object), undefined) : false;

            if (subResolved.resolved && value != null && (!emptyAsNull || !(typeof value == 'string') || !String.isNullOrEmpty(value as string))) {
                return value;
            }
            else {
                return parameters[1].resolve(expressionData, subResolved, /* running:*/ undefined);
            }

        }

        return undefined;
    }
}
interface LooseObject {
    [key: string]: any
}
export class toObject extends mfunction {

     /* override */ get minParameters(): number {
        return 0;
    }

     /* override */ get maxParameters(): number {
        return -1;
    }

     /* internal */ /* override */ invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        let source: Object | undefined = undefined;
        if ((parameters.length == 1)) {
            let subResolved: outputVariables = new outputVariables();
            source = parameters[0].resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
        }
        else {
            source = expressionData.contexts.peek();
        }

        if ((source != null)) {
            if (source instanceof namedSelection || source instanceof DictionaryObject || source instanceof selection) {
                let eo: LooseObject = {};
                //let eoColl = new List<KeyValuePair<string, Object>>();
                let dict = new DictionaryObject();
                if (source instanceof namedSelection) {
                    for (let strKey of source.getKeys()) {
                        dict.add(strKey.toString(), source.getValue(strKey) as object)
                    }
                }
                else if (source instanceof selection) {
                    let num = 0;
                    for (let strKey of source.toArray()) {
                        dict.add(num.toString(), strKey as object);
                        num++;
                    }
                }
                else {
                    dict = source as DictionaryObject;
                }

                for (let kvp of dict.getKeys()) {
                    //eoColl.add({ key: kvp, value: dict.getValue(kvp) as Object });
                    eo[kvp] = dict.getValue(kvp);
                }

                return (<LooseObject>(eo));
            }
            else {
                return source;
            }

        }

        return undefined;
    }
}

export class toObjectList extends mfunction {

     /* override */ get minParameters(): number {
        return 0;
    }

     /* override */ get maxParameters(): number {
        return -1;
    }

    invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        resolved.resolved = true;
        let source: Object | undefined = undefined;
        let func: toObject = new toObject();
        if ((parameters.length == 1)) {
            let subResolved: outputVariables = new outputVariables();
            source = parameters[0].resolve(expressionData, /* out */subResolved, /* running:*/ undefined);
        }
        else {
            source = expressionData.contexts.peek();
        }

        if ((source != null)) {
            if ((source as IEnumerable<any>).getEnumerator() !== undefined) {
                let newList: List<Object> = new List<Object>();
                for (let item of (<List<any>>(source)).toArray()) {
                    expressionData.contexts.push(item);
                    let subResolved: outputVariables = new outputVariables();
                    const aryExpressionPart: expressionPart[] = []
                    newList.add(func.invoke(expressionData, aryExpressionPart, /* out */subResolved, running) as Object);
                    expressionData.contexts.pop();
                }

                return newList;
            }

            return source;
        }

        return undefined;
    }
}